
% Sparsity-Promoting Optimal H_2 Controller for topology identification
%
% Written by: Sepideh Hassan-Moghaddam and Mihailo R. Jovanovic
%
% February 2016
%
% Proximal gradient with soft-thresholding

% Network Dynamics: qdot = -(Lp + Lx)*q + d
% 
% q - state of the network 
% d - white stochastic disturbance 
%
% Lp - graph Laplacian of the plant
% Lx = Ec*diag(x)*Ec' - graph Laplacian of the controller
% Ec - incidence matrix of the controller graph
% x  - vector of the edge weights
%
% Optimization Problem
%
% minimize <Qp,(Gp + Lx)^{-1}> + <R,Lx> + gamma*sum(x_i)
% 
% subject to Gp + E diag(x) E' > 0
% 
% Syntax:
%
% out = graphsp_proxBB(Lp,Ec,Q,R,gamma,opts)
%
% Inputs: (1) plant graph Laplacian Lp
%             incidence matrix of the controller graph Ec
%             state weight Q
%             controller weight R
%             sparsity-promoting parameter gamma
%
%         (2) options
%
%             opts.eps_dg    - tolerance on duality gap
%             opts.eps_pres   - tolerance on primal residual
%             opts.eps_res   - tolerance on dual residual
%             opts.maxit     - maximum number of IP iterations
%             opts.flagBB    - flag for BB step-size initialization

%
%             If opts argument is omitted, the default values are set to
%
%             opts.eps_dg    = 1.e-3
%             opts.eps_pres  = 1.e-5
%             opts.eps_dres  = 1.e-2
%             opts.maxit     = 1000
%             opts.flagBB    = 1
%
% Output:  out - output structure containing
%
%          out.x    - optimal vector of edge weights 
%          out.s    - number of proxBB iterations
%          out.dg   - duality gap
%          out.pres - primal residual
%          out.dres - dual residual
%          out.ss   - step-size
%          out.js   - number of iterations for line-search
%          out.iter_time - time of an inner iteration
%
% Additional information:
%
% http://www.umn.edu/~mihailo/software/graphsp/

function out = graphsp_proxBB(Lp,Ec,R,gamma,opts) 

% Initialization
if nargin < 4
    error('At least four input arguments are required.')
elseif nargin == 4
    opts = struct('maxit',1000,'eps_pres',1.e-5,'eps_dres',1.e-2,'eps_dg',1.e-3,'flagBB',1);
elseif nargin > 5
    error('Too many input arguments.')
end

% number of nodes
N = size(Ec,1);

% number of edges in the controller graph
% determines the number of optimization variables
m = size(Ec,2);

% Initialization
I = eye(N);

% Data preprocessing
maxit = opts.maxit;
eps_dg = opts.eps_dg;
eps_pres = opts.eps_pres;
eps_dres = opts.eps_dres;
flagBB = opts.flagBB;

% backtracking parameters
ssize = 1;
beta  = 0.5;

% "strengthened" plant graph Laplacian
Gp = Lp + (1/N)*ones(N,N);

% strengthened state weight 
% Qp = Q + (1/N)*ones(N,N) + Lp*R*Lp; 
% for Q = Qda = I - (1/N)*ones(N,N)
Qp = I + Lp*R*Lp; 

% % for Q = I - (1/N)*ones(N,N)
% Qp = I + Lp*R*Lp; 

% form the vector that appears in the objective function
qvec = sum(Ec.*(R*Ec))';

% initialize feedback gain
x = 0.5*ones(m,1); 

% compute invG
% G = Gp + Ec*sparse(1:m,1:m,x)*Ec';
invG = (Gp + Ec*sparse(1:m,1:m,x)*Ec')\I;

% objective function - smooth part
J = trace(invG*Qp) + (qvec)'*x;

% form dual variable from a primal one
Y = invG*Qp*invG;

% grad of J
gradJ = qvec - sum(Ec.*(Y*Ec))';

% Allocate memory for gamma-dependent output variables
out.x = zeros(m,length(gamma)); % vector of amplitudes (before polishing)
out.s = zeros(length(gamma)); % number of iterations
out.dg  = zeros(length(opts.maxit),1); % duality gap
out.dresp = zeros(length(opts.maxit),1); % dual residual plus
out.dresm = zeros(length(opts.maxit),1); % dual residual minus
out.pres = zeros(length(opts.maxit),1); % dual residual
out.js  = zeros(length(gamma),length(opts.maxit)); % number of inner iterations for back-tracking
out.ss  = zeros(length(gamma),length(opts.maxit)); % step-size


% print
fprintf('%s %s %s %s %s\n\n', ' d_gap', '    p_res', '   d_res_p', '  d_res_m', ' iter');   

% Use proximal gradient method to solve the gamma-parameterized problem
for s = 1:maxit,
    
    % save time of the iteration
    start1 = tic;
    
    % line search
    for j = 1:1000,
        
        % l1 norm with soft-threshholding operator
        xnew = sign(x - ssize*gradJ).*max(abs(x - ssize*gradJ) -...
        gamma*ssize,zeros(m,1));
        
        % Gnew
        Gnew = Gp + Ec*sparse(1:m,1:m,xnew)*Ec';
        eig_G = eig(Gnew);
        
        % check stability of the closed-loop system
        if min(eig_G) < 1.e-12
           
            ssize = ssize*beta;
            
        else
            
            %invGnew
            invGnew = (Gp + Ec*sparse(1:m,1:m,xnew)*Ec')\I;

            % objective function
            Jnew = trace(invGnew*Qp) + (qvec)'*xnew;

            % Gradient map
            GradJ = (1/ssize)*(x - xnew);
        
                % backtracking line search
                if Jnew <= J - ssize*gradJ'*GradJ + ...
                        0.5*ssize*(norm(GradJ)^2),

                    break;

                else

                    ssize = ssize*beta;

                end
            
        end
        
    end
    
    % record stepsize
    out.ss(s) = ssize;
    
    % record number of iterations
    out.js(s) = j;
   
    % form a dual variable 
    Ynew = invGnew*Qp*invGnew;
    
    % grad of J
    gradJnew = qvec - sum(Ec.*(Ynew*Ec))';
    
    % use BB stepsize initialization
    if flagBB == 1,
        
        % BB stepsize initialization
        ssize = (norm(xnew - x)^2) / ...
            ((x - xnew)'*(gradJ - gradJnew));
        
        if ssize < 0 || ssize == Inf,
            
            ssize = 1;
            % bbfailures(end+1) = iter;
            
        end
        
    else
        
        ssize = 1;
        
    end
    
    % xp, xm (x = xp - xm)
    indp = xnew>=0;
    xp = zeros(m,1);
    xm = zeros(m,1);
    xp(indp==1) = xnew(indp==1);
    xm(indp==0) = -xnew(indp==0); 

    % change of variable to guarantee dual feasibility
    temp = norm(gradJnew,inf);
    
    if (temp <= gamma)
    
        % dual feasible zp and zm
        zp = gamma + gradJnew;
        zm = gamma - gradJnew;
        
    else
        
        % beta
        beta1 = (gamma + 2*R(1,1))/(norm(gradJnew,inf)+ 2*R(1,1));
        
        % dual feasible Y (change of variable)
        Yprime = beta1 * Ynew + ((1-beta1)/N)*ones(N,N);
        
        % dual feasible zp and zm
        zp = qvec + gamma - sum(Ec.*(Yprime*Ec))';
        zm = -qvec + gamma + sum(Ec.*(Yprime*Ec))';
        
    end
    
    % duality gap
    eta = (xp'*zp + xm'*zm);
    
    % residual vector for the primal feasibility
    rpvec = xnew - xp + xm;

    % residual vectors for the dual feasibility
    rd_p_vec = gamma + qvec - sum(Ec.*(Ynew*Ec))' - zp; 
    rd_m_vec = gamma - qvec + sum(Ec.*(Ynew*Ec))' - zm;  
    
    % declare variables
    out.dg(s,:)  = eta;
    out.pres(s,:) = norm(rpvec);
    out.dresp(s,:) = norm(rd_p_vec);
    out.dresm(s,:) = norm(rd_m_vec);
    
    % variable used in the stopping criterion
    check = any( [eta > eps_dg, norm(rpvec) > eps_pres, norm(rd_p_vec) > eps_dres, norm(rd_m_vec) > eps_dres] );
    
    % check convergence

     if check == 0
        
        x = xnew;    

    % display
    disp([num2str(eta,'%6.1E'),'   ', ...
        num2str(norm(rpvec),'%6.1E'),'   ', ...
        num2str(norm(rd_p_vec),'%6.1E'),'   ', ...
        num2str(norm(rd_m_vec),'%6.1E'),'   ', ...
        num2str(s),'   '])
        
        out.iter_time(s) = toc(start1);
        
        break;
        
    end
    
    x = xnew;
    J = Jnew;
    gradJ = gradJnew;
    
    % record time of iteration
    out.iter_time(s) = toc(start1);
    
    if mod(s,10)==0
        
        % display
        disp([num2str(eta,'%6.1E'),'   ', ...
            num2str(norm(rpvec),'%6.1E'),'   ', ...
            num2str(norm(rd_p_vec),'%6.1E'),'   ', ...
            num2str(norm(rd_m_vec),'%6.1E'),'   ', ...
            num2str(s),'   '])
        
    end

    
end

% Record output data
out.s = s; % number of iterations
out.x = x; % vector of amplitudes


