
% Sparsity-Promoting Optimal H_2 Controller for topology identification
%
% Written by: Sepideh Hassan-Moghaddam and Mihailo R. Jovanovic
%
% February 2016
%
% sequential quadratic approximation using coordinate descent

% Network Dynamics: qdot = -(Lp + Lx)*q + d
% 
% q - state of the network 
% d - white stochastic disturbance 
%
% Lp - graph Laplacian of the plant
% Lx = Ec*diag(x)*Ec' - graph Laplacian of the controller
% Ec - incidence matrix of the controller graph
% x  - vector of the edge weights
%
% Optimization Problem
%
% minimize <Qp,(Gp + Lx)^{-1}> + <R,Lx> + gamma*sum(x_i)
% 
% subject to Gp + E diag(x) E' > 0
% 
% Syntax:
%
% out = graphsp_proxBB(Lp,Ec,Q,R,gamma,opts)
%
% Inputs: (1) plant graph Laplacian Lp
%             incidence matrix of the controller graph Ec
%             state weight Q
%             controller weight R
%             sparsity-promoting parameter gamma
%
%         (2) options
%
%             opts.eps_dg    - tolerance on duality gap
%             opts.eps_pres  - tolerance on primal residual
%             opts.eps_res   - tolerance on dual residual
%             opts.maxit     - maximum number of IP iterations
%             opts.flagBB    - flag for BB step-size initialization

%
%             If opts argument is omitted, the default values are set to
%
%             opts.eps_dg    = 1.e-3
%             opts.eps_pres  = 1.e-5
%             opts.eps_dres  = 1.e-2
%             opts.maxit     = 200
%
% Output:  out - output structure containing
%
%          out.x    - optimal vector of edge weights 
%          out.s    - number of proxBB iterations
%          out.dg   - duality gap
%          out.pres - primal residual
%          out.dres - dual residual
%          out.ss   - step-size
%          out.js   - number of iterations for line-search
%          out.iter_time - time of an inner iteration
%
% Additional information:
%
% http://www.umn.edu/~mihailo/software/graphsp/
%

function out = graphsp_proxN(Lp,Ec,R,gamma,opts)


% Initialization
if nargin < 4
    error('At least four input arguments are required.')
elseif nargin == 4
     opts = struct('maxit',200,'eps_pres',1.e-5,'eps_dres',1.e-2,'eps_dg',1.e-3);
elseif nargin > 5
    error('Too many input arguments.')
end

% number of nodes
N = size(Ec,1);

% number of edges in the controller graph
% determines the number of optimization variables
m = size(Ec,2);

% Initialization
I = eye(N);

% Data preprocessing
maxit = opts.maxit;
eps_dg = opts.eps_dg;
eps_pres = opts.eps_pres;
eps_dres = opts.eps_dres;

% Armijo rule data
alpha = 0.3;
beta  = 0.5;
tau = 1;

% tolerance
tol3 = 0.001;
 
% "strengthened" plant graph Laplacian
Gp = Lp + (1/N)*ones(N,N);

% strengthened state weight
% Qp = Q + (1/N)*ones(N,N) + Lp*R*Lp;
% for Q = Qda = I - (1/N)*ones(N,N)
Qp = I + Lp*R*Lp;

% form the vector that appears in the objective function
qvec = sum(Ec.*(R*Ec))';

% initialize feedback gain
x = 0.5*ones(m,1);

% invG
%G = Gp + Ec * diag(f) * Ec';
invG = (Gp + Ec*sparse(1:m,1:m,x)*Ec')\I;

% form dual variable from a primal one
Y = invG*Qp*invG;

% Hessian
P = (Ec'*Y*Ec).*(Ec'*invG*Ec);

% Gradient
gradJ = -sum(Ec.*(Y*Ec))' + qvec;


% objective function
J = trace(Qp*invG) + qvec'*x - ...
    trace(R*Gp) + gamma*sum(abs(x));

% Jnew
Jnew = 0;

% initialize eta
eta = 100;

    
% Allocate memory for gamma-dependent output variables
out.x = zeros(m,length(gamma)); % vector of amplitudes (before polishing)
out.s = zeros(length(gamma)); % number of iterations
out.dg  = zeros(length(opts.maxit),1); % duality gap
out.dresp = zeros(length(opts.maxit),1); % dual residual plus
out.dresm = zeros(length(opts.maxit),1); % dual residual minus
out.pres = zeros(length(opts.maxit),1); % dual residual
out.js  = zeros(length(gamma),length(opts.maxit)); % number of inner iterations for back-tracking
out.ss  = zeros(length(gamma),length(opts.maxit)); % step-size

% print
fprintf('%s %s %s %s %s\n\n', ' d_gap', '    p_res', '   d_res_p', '  d_res_m', ' iter'); 

% Use quadratic approximation method to solve the gamma-parameterized problem
for s = 1:maxit,
    
    % save time of the iteration
    start1 = tic;
    
    % initialize the search direction
    dx = zeros(m,1);
    
    % norm of dx
    normdx = 0;
    
    % sum of differences in dx (mu's)
    musum = 0;

    %Pdx = P  * dx;
    Pdx = zeros(m,1);

    % active set strategy
    allind = 1:m;
    postindex = find(abs(gradJ) < gamma - .0001*gamma);
    
    % find zero elements in x
    zerox = find(x==0);
    fixedindex = intersect(zerox, postindex);
    
    % active set
    index = setdiff(allind,fixedindex);

    
    % going through the active set and update the directions
    for ind = 1: 5*m
            
            % going through the active set
            for iin = 1:size(index,2),
                
                    % ith element of the active set
                    i = index(iin);
                    
                    % ith diagonal element of P
                    a = P(i,i);
                    
                    % b/a
                    b_over_a = (Pdx(i) + gradJ(i))/a;
                    
                    % c
                    c = x(i) + dx(i);
                    
                    % c - b/a
                    cba = c -  b_over_a;
                    
                    % find the update mu for the search direction dx
                    mu = -c + sign(cba).*max(abs(cba) - gamma/a,0);
                    
                    % update dx
                    dx(i) = dx(i) + mu;
                    
                    % norm of dx
                    normdx = normdx + abs(dx(i));
                    
                    % sum of mu's
                    musum = musum + abs(mu);
                    
                    % P times dx
                    Pdx = Pdx + mu * P(:,i);
                

            end
                
            % stopping criterion of the inner iteration
            % measures how big are the changes in each direction compared to norm of the vector of edge weights.
                if ( musum <= normdx * tol3)
                     break;
                end
                
            % reset    
            musum = 0;
            normdx = 0;
 
    end




    % stability of the closed-loop system + Armijo rule
    for j = 1:1000,
        
        % update x
        xnew = x + tau*dx;
        
        % update G
        Gnew = Gp + Ec*sparse(1:m,1:m,xnew)*Ec';
        
        % eigen-values of Gnew
        eig_G = eig(Gnew);
        
        % check stability
        if min(eig_G) < 1.e-12
           
            tau = tau*beta;
            
        else
 
            % invGnew
            invGnew = Gnew\I;

            % form dual variable from a primal one
            Ynew = invGnew*Qp*invGnew;

            % objective function
            Jnew = trace(Qp*invGnew) + qvec'*xnew - ...
                   trace(R*Gp) + gamma*sum(abs(xnew));
            
            % Armijo rule
            if Jnew <= J + tau*alpha*( gradJ'*dx + ...
               gamma*( sum(abs(x + dx)) - sum(abs(x)) ) );

                break;

            end

            tau = tau*beta;
        
        end
        
    end
    
    
    % record number of iterations
    out.js(s) = j;
    
    % record stepsize
    out.ss(s) = tau;
    
    
    % update J 
    J = Jnew;
    
    % update x
    x = xnew;
    
    % update invG
    invG = invGnew;
    
    % update Y
    Y = Ynew;

    % update P
    P = (Ec'*Y*Ec).*(Ec'*invG*Ec);
    
    % gradient
    gradJ = -sum(Ec.*(Y*Ec))' + qvec;
   
    % reset tau
    tau = 1;
    
    % record time of iteration
    out.iter_time(s) = toc(start1);
      
    % xp, xm (x = xp - xm)
    indp = xnew >= 0;
    xp = zeros(m,1);
    xm = zeros(m,1);
    xp(indp==1) = xnew(indp==1);
    xm(indp==0) = -xnew(indp==0); 
    
    
    % change of variable to guarantee dual feasibility
    temp = norm(gradJ,inf);
    
    if (temp <= gamma)
    
        % dual feasible z
        zp = gamma + gradJ;
        zm = gamma - gradJ;
        
    else
        
        % beta
        beta1 = (gamma + 2*R(1,1))/(norm(gradJ,inf)+ 2*R(1,1));
        
        % dual feasible Y (change of variable)
        Yprime = beta1 * Ynew + ((1-beta1)/N)*ones(N,N);
        
        % dual feasible z
        zp = qvec + gamma - sum(Ec.*(Yprime*Ec))';
        zm = -qvec + gamma + sum(Ec.*(Yprime*Ec))';
        
    end
    
    % duality gap
    etanew = (xp'*zp + xm'*zm);
    
    % residual vector for the primal feasibility
    rpvec = xnew - xp + xm;

    % residual vectors for the dual feasibility
    rd_p_vec = gamma + qvec - sum(Ec.*(Ynew*Ec))' - zp; 
    rd_m_vec = gamma - qvec + sum(Ec.*(Ynew*Ec))' - zm;  
    
    % declare variables
    out.dg(s,:)  = etanew;
    out.pres(s,:) = norm(rpvec);
    out.dresp(s,:) = norm(rd_p_vec);
    out.dresm(s,:) = norm(rd_m_vec);
    
    % variable used in the stopping criterion
    check = any( [etanew > eps_dg, norm(rpvec) > eps_pres, norm(rd_p_vec) > eps_dres, norm(rd_m_vec) > eps_dres] );

    if check==0
        
        % display duality gap, dual residual, primal residual
         disp([num2str(eta,'%6.1E'),'   ', ...
        num2str(norm(rpvec),'%6.1E'),'   ', ...
        num2str(norm(rd_p_vec),'%6.1E'),'   ', ...
        num2str(norm(rd_m_vec),'%6.1E'),'   ', ...
        num2str(s),'   '])
        

        break;
        
    end
    
    if abs(eta-etanew)<1.e-8
        
      % display duality gap, dual residual, primal residual
         disp([num2str(eta,'%6.1E'),'   ', ...
        num2str(norm(rpvec),'%6.1E'),'   ', ...
        num2str(norm(rd_p_vec),'%6.1E'),'   ', ...
        num2str(norm(rd_m_vec),'%6.1E'),'   ', ...
        num2str(s),'   '])
         
        break;
    end
    
    % update eta
    eta = etanew;
    
    if mod(s,10)==0
     
        % display duality gap, dual residual, primal residual
     disp([num2str(eta,'%6.1E'),'   ', ...
        num2str(norm(rpvec),'%6.1E'),'   ', ...
        num2str(norm(rd_p_vec),'%6.1E'),'   ', ...
        num2str(norm(rd_m_vec),'%6.1E'),'   ', ...
        num2str(s),'   '])
    end
    
end
% Record output data
out.s = s; % number of iterations
out.x = x; % vector of amplitudes